'''This file contains required class and methods for exercise 2'''
import pandas as pd
from analyticsTestNN1 import helpers


class Ops:
    '''This class provides methods to to be used in exercise 3 on DataFrames'''

    @staticmethod
    def remove_null_values(data: pd.DataFrame) -> pd.DataFrame:
        '''
        Removes rows that contain a null value.
        A null value (in the context of this func) is one of the following:
            i. nan (np.NaN)
            ii. None
            iii. ""

        Args:
            data (pd.DataFrame) : The pandas dataframe

        Returns:
            pd.DataFrame : a dataframe without null values

        Note: Can be implemented using pandas functions:

            return data.dropna()

            or

            return data[data.apply(lambda x: not pd.isnull(x).any(), axis = 1)]
        '''
        if data.empty:
            return data

        boolean_vals = []
        for row in data.values:
            boolean_vals.append(not helpers._isnull(row))
        return data[boolean_vals]

    @staticmethod
    def rotate(
        data: pd.DataFrame,
        xx: str, yy: str,
            scalars: str) -> pd.DataFrame:
        '''
        Implements the pivot functionality of the pandas DataFrames

        Args:
            data (pd.DataFrame): The pandas dataframe to be rotated
            xx (str) : The column to be the x-axis of rotated dataframe
            yy (str) : The column to be the y-axis of rotated dataframe
            scalars(str) : The column that will populate the dataframe

        Returns:
            pd.DataFrame : The rotated (pivoted) dataframe

        Raises:
            Exception: ValueError Index contains duplicate entires

        Note: Can be implemented using pandas functions:

            return data.pivot(index=xx, columns=yy, values=scalars)
        '''

        data = Ops.remove_null_values(data)

        if data.empty:
            return data

        dict_of_dicts = {}

        # Iterate over each row and unpack values
        for _, (x, y, z) in data[[xx, yy, scalars]].iterrows():

            if y in dict_of_dicts:
                # if x is already a key, then there are duplicate values
                if x in dict_of_dicts[y]:
                    raise ValueError("Value Error: \
                        Index contains duplicate entries")
                else:
                    dict_of_dicts[y][x] = z
            else:
                dict_of_dicts[y] = {}
                dict_of_dicts[y][x] = z

        return_df = pd.DataFrame(data=dict_of_dicts)
        return_df.index.name = xx
        return_df.columns.name = yy
        return return_df



'''
This module contains helper functions that simplify the code in analytics.py
'''
import numpy as np


def _isnull(row: np.ndarray) -> bool:
    '''
    Returns True if given numpy array contains a null value 
    Null value of form:
        i. nan (np.NaN)
        ii. None
        iii. ""
    '''
    for val in row:
        if val:
            if isinstance(val, float):
                if np.isnan(val):
                    return True
        else:
            return True
    return False



from analyticsTestNN1 import analytics
import pandas as pd
import numpy as np


def test_remove_null_values_empty():
    '''Test where an empty df is passed to function'''
    df_test = pd.DataFrame()
    result_df = analytics.Ops.remove_null_values(df_test)

    pd.testing.assert_frame_equal(df_test, result_df)


def test_remove_null_values_no_nulls():
    '''Test where no nulls in df'''
    df_test = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                           columns=['A', 'B', 'C'])
    result_df = analytics.Ops.remove_null_values(df_test)

    pd.testing.assert_frame_equal(df_test, result_df)


def test_remove_null_values_one_nan():
    '''Test where df contains a single nan value'''
    df_test = pd.DataFrame([[1, np.nan, 3], [4, 5, 6]],
                           columns=['A', 'B', 'C'])

    # Drop first row containing nan value
    df_expected = df_test.iloc[1:]
    result_df = analytics.Ops.remove_null_values(df_test)

    pd.testing.assert_frame_equal(df_expected, result_df)


def test_remove_null_values_one_None():
    '''Test where df contains a single None value'''
    df_test = pd.DataFrame([['a', 'b', 'c'], ['d', 'e', 'f'], [4, 5, None]],
                           columns=['A', 'B', 'C'])

    # Drop last row containing None value
    df_expected = df_test.iloc[:-1]
    result_df = analytics.Ops.remove_null_values(df_test)

    pd.testing.assert_frame_equal(df_expected, result_df)


def test_remove_null_values_all_rows():
    '''Test where df contains a null value in each row'''
    df_test = pd.DataFrame([[np.nan, 'b'], [np.nan, 'f'], [5, None]],
                           columns=['A', 'B'])

    result_df = analytics.Ops.remove_null_values(df_test)

    assert result_df.empty


def test_remove_null_values_given_example():
    '''Test for the given example in exercise3.py'''
    df_test = pd.DataFrame([['a', 'AXA', 13, 'djasd8'],
                            ['b', 'BZB', 123.12000, '7123hy'],
                            ['a', 'CYC', 3.40000, 'h6as7d'],
                            ['a', 'BZB', 4.41332, 'Naisd871a'],
                            ['b', 'AXA', 54.00000, 'dashd77'],
                            ['b', 'CYC', 6.12000, 'mdas7gg'],
                            ['b', None, 612.10000, 'masf7gg'],
                            ['a', 'AXA', np.nan, 'jdasd765ad']],
                           columns=['A', 'B', 'C', 'D'])

    result_df = analytics.Ops.remove_null_values(df_test)

    df_expected = pd.DataFrame([['a', 'AXA', 13, 'djasd8'],
                                ['b', 'BZB', 123.12000, '7123hy'],
                                ['a', 'CYC', 3.40000, 'h6as7d'],
                                ['a', 'BZB', 4.41332, 'Naisd871a'],
                                ['b', 'AXA', 54.00000, 'dashd77'],
                                ['b', 'CYC', 6.12000, 'mdas7gg']],
                               columns=['A', 'B', 'C', 'D'])

    assert df_expected.equals(result_df)



from analyticsTestNN1 import analytics
import pandas as pd


def test_rotate_empty():
    '''Test where an empty df is passed to function'''
    df_test = pd.DataFrame(columns=['A', 'B', 'C'])
    result_df = analytics.Ops.rotate(df_test, xx='A', yy='B', scalars='C')

    pd.testing.assert_frame_equal(df_test, result_df)


def test_rotate_example():
    '''Test with the example given'''
    df_test = pd.DataFrame([['a', 'AXA', 13, 'djasd8'],
                            ['b', 'BZB', 123.12000, '7123hy'],
                            ['a', 'CYC', 3.40000, 'h6as7d'],
                            ['a', 'BZB', 4.41332, 'Naisd871a'],
                            ['b', 'AXA', 54.00000, 'dashd77'],
                            ['b', 'CYC', 6.12000, 'mdas7gg']],
                           columns=['A', 'B', 'C', 'D'])

    result_df = analytics.Ops.rotate(df_test, xx='A', yy='B', scalars='C')

    df_expected = pd.DataFrame([[13, 4.41332, 3.40000],
                                [54.00000, 123.12000, 6.12000]],
                               columns=['AXA', 'BZB', 'CYC'],
                               index=['a', 'b'])
    df_expected.index.name = 'A'
    df_expected.columns.name = 'B'

    pd.testing.assert_frame_equal(df_expected, result_df)



from setuptools import setup

setup(
    name="analyticsTestNN1",
    version="1.0.0",
    description="Testing out the analytics package",
    author="Testing@testing",
    license="MIT",
    packages=["analyticsTestNN1"],
    install_requires=["pandas"],
)


from analyticsTestNN1 import helpers
import numpy as np
import pandas as pd


def test_no_null():

    # Simulating how numpy array is extracted from pandas
    np_array = pd.Series([1, 2, 3, '4', 'xyx']).values
    assert not helpers._isnull(np_array)


def test_nan():

    np_array = pd.Series([1, 2, np.nan, '4', 'xyx']).values
    assert helpers._isnull(np_array)


def test_none():

    np_array = pd.Series([1, None, 3, '4', 'xyx']).values
    assert helpers._isnull(np_array)


def test_NaT():

    np_array = pd.Series([1, pd.NaT, 3, '4', 'xyx']).values
    assert not helpers._isnull(np_array)


def test_empty_string():
    '''Testing case where empty string exists, returns True'''

    np_array = pd.Series([1, 2, 3, '', 'xyx']).values
    assert helpers._isnull(np_array)
